#include "terminal.h"
#include "ctype.h"
#include "kprintf.h"
#include "string.h"

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

static inline uint8_t vga_entry_color(enum vga_color fg, enum vga_color bg)
{
	return fg | bg << 4;
}

static inline uint16_t vga_entry(unsigned char uc, uint8_t color)
{
	return (uint16_t)uc | (uint16_t)color << 8;
}

#define VGA_WIDTH 80
#define VGA_HEIGHT 25

size_t terminal_row;
size_t terminal_column;
uint8_t terminal_color;
uint16_t *terminal_buffer;
void terminal_initialize(void)
{
	terminal_row = 0;
	terminal_column = 0;
	terminal_color = vga_entry_color(VGA_COLOR_LIGHT_GREY, VGA_COLOR_BLACK);
	terminal_buffer = (uint16_t *)0xB8000;
	for (size_t y = 0; y < VGA_HEIGHT; y++) {
		for (size_t x = 0; x < VGA_WIDTH; x++) {
			const size_t index = y * VGA_WIDTH + x;
			terminal_buffer[index] = vga_entry(' ', terminal_color);
		}
	}
}

void terminal_setcolor(uint8_t color)
{
	terminal_color = color;
}

void terminal_putentryat(char c, uint8_t color, size_t x, size_t y)
{
	const size_t index = y * VGA_WIDTH + x;
	terminal_buffer[index] = vga_entry(c, color);
}

static void terminal_scroll(void)
{
	terminal_row--;
	for (size_t i = 0; i < VGA_WIDTH * (VGA_HEIGHT - 1); i++)
		terminal_buffer[i] = terminal_buffer[i + VGA_WIDTH];
	for (size_t i = 0; i < VGA_WIDTH; i++)
		terminal_putentryat(' ', terminal_color, i, VGA_HEIGHT - 1);
}

static void terminal_new_line(void)
{
	terminal_column = 0;
	if (++terminal_row == VGA_HEIGHT)
		terminal_scroll();
}

int terminal_putchar(char c)
{
	if (c == '\r')
		terminal_column = 0;
	else if (c == '\n')
		terminal_new_line();
	if (!isprint(c))
		return 1;
	terminal_putentryat(c, terminal_color, terminal_column, terminal_row);
	if (++terminal_column == VGA_WIDTH)
		terminal_new_line();
	return 1;
}

int terminal_write(const char *data, size_t size)
{
	size_t i;
	for (i = 0; i < size; i++)
		terminal_putchar(data[i]);
	return (int)i;
}

int terminal_writestring(const char *data)
{
	size_t len = strlen(data);
	terminal_write(data, len);
	return len;
}

int terminal_writelong(long n)
{
	long div = 10;
	int rv = 0;
	if (n < 0) {
		rv += terminal_putchar('-');
		n *= -1;
	}
	div = 1;
	while (div <= n / 10)
		div *= 10;
	while (div > 0) {
		rv += terminal_putchar('0' + n / div % 10);
		div /= 10;
	}
	return rv;
}
